/* LD_PRELOADable library to stop Mozilla from finding plugins.
 * Adam Sampson <azz@us-lot.org>
 *
 * This may only work on Solaris.
 */

#include <sys/types.h>
#include <dirent.h>

#include <dlfcn.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

typedef struct {
	DIR *dir;
	char *name;
} dirname;
#define MAX_DIRS 256
dirname dirs[MAX_DIRS];

typedef struct allowed {
	char *name;
	struct allowed *next;
} allowed;
allowed *allowed_list = NULL;

static void die(const char *s) {
	fprintf(stderr, "killplugins failed: %s\n", s);
	abort();
}

static void add_allowed(const char *name) {
	allowed *a = malloc(sizeof *a);
	if (a == NULL)
		die("out of memory");

	a->name = strdup(name);
	a->next = allowed_list;
	allowed_list = a;
}

static void init() {
	int i;
	const char *home = getenv("HOME");
	static int done = 0;
	if (done)
		return;

	for (i = 0; i < MAX_DIRS; i++)
		dirs[i].dir = NULL;

	add_allowed(".");
	add_allowed("..");

	if (home != NULL) {
		FILE *f;
		char *s = malloc(strlen(home) + 18);
		if (s == NULL)
			die("out of memory");

		strcpy(s, home);
		strcat(s, "/.mozilla-plugins");
		f = fopen(s, "r");
		free(s);

		if (f != NULL) {
			char buf[FILENAME_MAX];
			while (fgets(buf, sizeof buf, f) != NULL) {
				buf[strlen(buf) - 1] = '\0';
				add_allowed(buf);
			}
			fclose(f);
		}
	}

	done = 1;
}

DIR *opendir(const char *dirname) {
	int i;
	DIR *d;
	static DIR *(*real)(const char *) = NULL;
	if (real == NULL)
		real = dlsym(RTLD_NEXT, "opendir");

	d = real(dirname);
	if (d == NULL)
		return NULL;

	init();
	for (i = 0; i < MAX_DIRS; i++) {
		if (dirs[i].dir != NULL)
			continue;

		dirs[i].dir = d;
		dirs[i].name = strdup(dirname);
		break;
	}

	return d;
}

/* FIXME might also want to wrap readdir_r */
struct dirent *readdir(DIR *dirp) {
	int i;
	static struct dirent *(*real)(DIR *) = NULL;
	if (real == NULL)
		real = dlsym(RTLD_NEXT, "readdir");

	init();
	for (i = 0; i < MAX_DIRS; i++) {
		if (dirs[i].dir != dirp)
			continue;

		if (strstr(dirs[i].name, "/plugins") == NULL)
			continue;

		while (1) {
			allowed *a;
			struct dirent *de = real(dirp);
			if (de == NULL)
				return NULL;

			for (a = allowed_list; a != NULL; a = a->next) {
				if (strcmp(a->name, de->d_name) == 0)
					return de;
			}
		} 
	}

	return real(dirp);
}

int closedir(DIR *dirp) {
	int i;
	static int (*real)(DIR *) = NULL;
	if (real == NULL)
		real = dlsym(RTLD_NEXT, "closedir");

	init();
	for (i = 0; i < MAX_DIRS; i++) {
		if (dirs[i].dir != dirp)
			continue;

		dirs[i].dir = NULL;
		free(dirs[i].name);
		break;
	}

	return real(dirp);
}

